#include "system.h"

#include <ctype.h>

#include "SystemColour.h"
#include "ColouredLabel.h"

#define ESCAPE '\033'

#ifdef USE_GTK
void coloured_label_redraw(GtkWidget * widget, GdkEvent * event, gpointer data) {
  ColouredLabel * cl = (ColouredLabel *)data;
  cl->redraw();
}
#endif

ColouredLabel::ColouredLabel(Connection * c) {
  conn = c;

  widget = NULL;
  widget_text = strdup("");
}

ColouredLabel::ColouredLabel(Connection * c, char * text) {
  conn = c;

  widget = NULL;
  widget_text = strdup("");
#ifdef USE_GTK
  setText(text);
#endif
}

ColouredLabel::~ColouredLabel() {
  free(widget_text);
}

void ColouredLabel::init() {

  papaya_colour = false;
  bold = false;
  background_color_default = true;
  foreground_color_default = true;
  flashing = false;
  span = false;

  // Initialise color and background_color
  fg = conn->queryPreferences()->getDefaultForegroundColour();
  bg = conn->queryPreferences()->getDefaultBackgroundColour();

  struct colour_table * table = conn->queryPreferences()->getColourTable();

  color_map = gdk_colormap_get_system();

  color.red = table[fg].red;
  color.green = table[fg].green;
  color.blue = table[fg].blue;

  background_color.red = table[bg].red;
  background_color.green = table[bg].green;
  background_color.blue = table[bg].blue;

  gdk_color_alloc(color_map, &color);
  gdk_color_alloc(color_map, &background_color);
}

#ifdef USE_GTK_2
void ColouredLabel::createWidget() {
  init();

  widget = gtk_label_new("");
}
#else
void ColouredLabel::createWidget() {
  init();

  widget = gtk_drawing_area_new();

  old_width = old_height = 0;

  gtk_signal_connect(GTK_OBJECT(widget), "expose_event",
		     GTK_SIGNAL_FUNC(coloured_label_redraw),
		     this);
  
  gtk_signal_connect(GTK_OBJECT(widget), "configure_event",
  		     GTK_SIGNAL_FUNC(coloured_label_redraw),
  		     this);
}
#endif

// Text string includes ANSI colour codes.

void ColouredLabel::setText(char * text) {

#ifdef USE_GTK_2
  char inter[16384];
  char output[16384];
  char * inter_ptr;
  char * output_ptr;
  char * pc = text;
  char * last_marker = text;

  inter_ptr = inter;
  inter[0] = '\0';

  output_ptr = output;
  output[0] = '\0';

  while (*pc != '\0') {
	  if (*pc == '>' || *pc == '<' || *pc == '&') {
		  *inter_ptr++ = '&';
		  if (*pc == '>') {
			  *inter_ptr++ = 'g';
			  *inter_ptr++ = 't';
		  }
		  else if (*pc == '<') {
			  *inter_ptr++ = 'l';
			  *inter_ptr++ = 't';
		  }
		  else {
			  *inter_ptr++ = 'a';
			  *inter_ptr++ = 'm';
			  *inter_ptr++ = 'p';
		  }
		  *inter_ptr++ = ';';
	  } else {
		  *inter_ptr++ = *pc;
	  }
	  pc++;
  }

  *inter_ptr = '\0';

  pc = inter;
  last_marker = pc;

  while ((pc = strchr(pc, ESCAPE))) {
    *pc = '\0';
    
    if (last_marker != '\0') {
      printPrompt(last_marker, output_ptr);
      output_ptr = output + strlen(output);
    }

    pc = parseVTCode(pc);
    last_marker = pc;
  }

  if (last_marker != '\0')
    printPrompt(last_marker, output_ptr);

  if (span)
    strcat(output_ptr, "</span>");

  printf("Setting markup.\n");
  gtk_label_set_markup(GTK_LABEL(widget), output);
  printf ("Done.\n");

  span = false;

#else
  if (widget_text)
    free(widget_text);
  widget_text = strdup(text);

  redraw();
#endif
}

GtkWidget * ColouredLabel::getWidget() {
  return widget;
}

#ifdef USE_GTK
void ColouredLabel::redraw() {

  int y_offset;         // Determined later
  int w_width;          // Window width
  int w_height;         // Window height

  int f_left_bearing;   // Extent of the string left of the drawing origin
  int f_right_bearing;  // Extent of the string right of the drawing origin
  int f_width;          // Total width of the string in this font
  int f_ascent;         // Ascent above the baseline
  int f_descent;        // Decent below the baseline
  GdkFont * font = conn->queryPreferences()->getPrompt();

  // Initialise all variables involved in drawing the prompt.
  prompt_index = 0;
  x_offset = 1; // 1 pixel border

  if (!widget_text)
    return;

  if (!widget)
    return;

  gdk_window_get_size(widget->window, &w_width, &w_height);

  // Do the background colour for the drawing area.

  char * start_pc = strdup(widget_text);
  char * pc = start_pc;
  char * last_marker = start_pc;

  gdk_string_extents(font, widget_text, &f_left_bearing, &f_right_bearing, &f_width, &f_ascent, &f_descent);
  y_offset = f_ascent + (w_height - (f_ascent + f_descent)) / 2;

  plain_text[0] = '\0';

  gdk_window_clear(widget->window);

  // Start parsing text...  Borrow some code from the ColouredLabel implementation.
  while ((pc = strchr(pc, ESCAPE))) {
    *pc = '\0';

    if (last_marker != '\0') {
      printPrompt(last_marker, y_offset);
      strcat(plain_text, last_marker);
    }

    pc = parseVTCode(pc);
    last_marker = pc;
  }

  if (last_marker != '\0') {
    printPrompt(last_marker, y_offset);
    strcat(plain_text, last_marker);
  }
  
  free(start_pc);

  if (!font)
    return;
  int width = gdk_string_width(font, plain_text) + 2;   // 1 pixel borders
  int height = gdk_string_height(font, plain_text) + 2;

  // Setting the drawing area size generates an expose event, sigh.
  // But removing and re-adding the expose_event doesn't help.
  // Don't resize the widget unless we actually need to.

  if (width == old_width && height == old_height)
    return;

  gtk_widget_hide(GTK_WIDGET(widget));
  gtk_drawing_area_size(GTK_DRAWING_AREA(widget), width, height);
  gtk_widget_show(GTK_WIDGET(widget));

  old_width = width;
  old_height = height;
}
#endif

void ColouredLabel::printPrompt(char * text, char * output) {

  if (!text) {
    printf ("FIXME: ColouredLabel::printPrompt given NULL string.\n");
    return;
  }

  if (span) {
    strcat(output, "</span>");
    output += strlen("</span>");
  }

  char foreground[16384];
  char background[16384];

  foreground[0] = '\0';
  background[0] = '\0';

  if (!foreground_color_default || papaya_colour)
    snprintf(foreground, 16384, " foreground=\"#%02x%02x%02x\"",
	     color.red * 255 / 65535 , color.green * 255 / 65535 , color.blue * 255 / 65535);
  
  if (!background_color_default)
    snprintf(background, 16384, " foreground=\"#%02x%02x%02x\"",
	     background_color.red * 255 / 65535 , background_color.green * 255 / 65535 , background_color.blue * 255 / 65535);

  snprintf(output, 16384, "<span font_desc=\"%s\"%s%s>%s",
	   conn->queryPreferences()->getPromptFont(),
	   foreground, background, text);
  
  span = true;
}

void ColouredLabel::printPrompt(char * string, int y_offset) {

  // Create a graphics context and set the colours in it.

  GdkGC * gc = gdk_gc_new(widget->window);
  if (!gc)
    return;

  gdk_gc_copy(gc, widget->style->white_gc);

  gdk_gc_set_foreground(gc, &color);
  gdk_gc_set_background(gc, &background_color);

  // What's the function calls to draw text on a drawing area?
  GdkFont * font = conn->queryPreferences()->getPrompt();

  gdk_draw_string(widget->window,
		  font,
		  gc,
		  x_offset,
		  y_offset,
		  string);

  x_offset += gdk_string_width(font, string);

  gdk_gc_destroy(gc);
}

char * ColouredLabel::parseVTCode(char * code) {
  char *pc = code+1;

  switch (*pc) {
  case '[':
    pc = parseBracketVTCode(pc);
    break;

  default:
    pc = parsePlainVTCode(pc);
    break;
  }

  pc++;
  return pc;
}

char * ColouredLabel::parseBracketVTCode(char * pc) {

  char * orig = pc;
  
  while (!isalpha(*pc))
    pc++;
  
  switch (*pc) {
    
  case 'm': // Colour code
    setColour(orig);
    break;
    
  case 'p': // Internal Papaya colour code.
    setPapayaColour(orig);
    break;
    
  case 'q': // Reset Papaya colour.
    resetPapayaColour();
    
  }
  
  return pc;
}

char * ColouredLabel::parsePlainVTCode(char * pc) {
  while (!isalpha(*pc))
    pc++;

  return pc;
}
void ColouredLabel::setPapayaColour(char * buf) {

  int red;
  int green;
  int blue;

  // allocate color

  char * pc;

  pc = strchr(buf, 'p');
  if (!pc)
    return;

  if (pc - buf != 12) { // [rrr;ggg;bbb
    printf("Dud Papaya Colour code.\n");
    return;
  }

  *pc = '\0';
  if (sscanf(buf, "[%d;%d;%d", &red, &green, &blue) != 3) {
    printf("PapayaColour: unable to sscanf.\n");
    return;
  }

  papaya_colour = TRUE;

  color.red = red * 0xffff / 255;
  color.green = green * 0xffff / 255;
  color.blue = blue * 0xffff / 255;

  gdk_color_alloc(color_map, &color);
}


void ColouredLabel::resetPapayaColour() {
  papaya_colour = FALSE;

  if (bold) {
    struct colour_table * table = conn->queryPreferences()->getColourTable();

    color.red = table[fg+8].red;
    color.green = table[fg+8].green;
    color.blue = table[fg+8].blue;

    background_color.red = table[bg + 8].red;
    background_color.green = table[bg + 8].green;
    background_color.blue = table[bg + 8].blue;

    gdk_color_alloc(color_map, &color);
    gdk_color_alloc(color_map, &background_color);
  } else {
    struct colour_table * table = conn->queryPreferences()->getColourTable();

    color.red = table[fg].red;
    color.green = table[fg].green;
    color.blue = table[fg].blue;

    background_color.red = table[bg].red;
    background_color.green = table[bg].green;
    background_color.blue = table[bg].blue;

    gdk_color_alloc(color_map, &color);
    gdk_color_alloc(color_map, &background_color);
  }
}

void ColouredLabel::setColour(char * buf) {
  char * pc = buf;
  char buf2[10];
  int pos = 0;

  buf2[0] = '\0';
  
  papaya_colour = FALSE;

  while (1) {

    // If we haven't got a [ or ; indicating more codes, leave the loop.
    if (!(*pc == ';' || *pc == '['))
      break;

    // Rip out the bit until the next non-alphanumeric character.
    pos = 0;

    buf2[pos] = *pc++;

    while (1) {
      if (isalpha(*pc) || *pc == ';')
        break;

      buf2[pos++] = *pc++;
    }
    
    buf2[pos] = '\0';
    setActualColour(buf2);
  }
}

void ColouredLabel::setActualColour(char * buf) {
  if (!conn->queryPreferences()->getColour())
    return;

  if (buf[0] == '0') {

    // Set everything to defaults.
    background_color_default = TRUE;
    foreground_color_default = TRUE;

    bold = FALSE;
    flashing = FALSE;
    return;
  }

  if (buf[0] == '1') {
    bold = TRUE;

    struct colour_table * table = conn->queryPreferences()->getColourTable();

    color.red = table[fg+8].red;
    color.green = table[fg+8].green;
    color.blue = table[fg+8].blue;

    background_color.red = table[bg + 8].red;
    background_color.green = table[bg + 8].green;
    background_color.blue = table[bg + 8].blue;

    gdk_color_alloc(color_map, &color);
    gdk_color_alloc(color_map, &background_color);

    return;
  }

  if (buf[0] == '5') {
    flashing = TRUE;
    return;
  }

  /* Look up which colour we're to use. */

  /* buf[2] should contain the colour. */

  if (strlen(buf) >= 2) {
    for (int i = 0; i < 8; i++) {

      struct colour_table * table = conn->queryPreferences()->getColourTable();

      if (table[i].colour_character == buf[1]) {

        if (buf[0] == '3') {
          foreground_color_default = FALSE;

          if (bold) {
            color.red = table[i+8].red;
            color.green = table[i+8].green;
            color.blue = table[i+8].blue;
          } else {
            color.red = table[i].red;
            color.green = table[i].green;
            color.blue = table[i].blue;
          }
          gdk_color_alloc(color_map, &color);
          fg = i;
          return;
        }
        
        if (buf[0] == '4') {
          background_color_default = FALSE;

          if (bold) {
            background_color.red = table[i+8].red;
            background_color.green = table[i+8].green;
            background_color.blue = table[i+8].blue;
          } else {
            background_color.red = table[i].red;
            background_color.green = table[i].green;
            background_color.blue = table[i].blue;
          }
          gdk_color_alloc(color_map, &background_color);
          bg = i;
          return;
        }
        
        return;

      }
    }
  }
}
