#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "system.h"
#include "EntityHandler.h"
#include "Event.h"
#include "Socks.h"
#include "SocksFive.h"
#include "PluginHandler.h"
#include "BaseWindow.h"
#include "Connection.h"
#include "Prefs.h"
#include "SystemTriggerEntity.h"

#ifdef USE_GTK_2
#include "GTKTwoVT.h"
#else
#include "PapayaVT.h"
#endif

#include "HTTPSocket.h"
#include "SquidSocket.h"

extern PluginHandler * phandler;
extern int read_socket_timeout(gpointer data);
extern BaseWindow * mainWindow;
extern EntityHandler * entities;

// The global preferences data structure.
extern Prefs * globalPreferences;

// Function pointer for plugins to over-write as needed
void (*keepalive_send)(Connection *) = &Connection::keepalive_send;

MUD * Connection::lastMUD = NULL;


void Connection::init() {

  loginTrigger = NULL;
  passwordTrigger = NULL;

  name = NULL;
  connected = 0;
  copyover_event = TRUE;
  charMode = FALSE;
  deleted = FALSE;
  isclosed = false;

  next = (Connection *)NULL;

#ifdef USE_GTK_2
  vt = new GTKTwoVT(this);
#else
  vt = new PapayaVT(this);
#endif

  if (queryPreferences()->getHttpConnect()) {
    s = new SquidSocket(this);
  } else {
    if (queryPreferences()->getTurfHttpd()) {
      s = new HTTPSocket(this);
    } else {
      if (queryPreferences()->getSocks()) {
	if (queryPreferences()->getSocks4()) {
	  s = new Socks(this);
	} else {
	  s = new SocksFive(this);
	}
      } else {
	s = new Socket(this);
      }
    }
  }

  timeout = 0;

  history = new History(queryPreferences()->getHistoryUseLimit()
			? queryPreferences()->getHistoryLimit() : 0,
			queryPreferences()->getHistoryIsCircular());
}

Connection::Connection(MUD * m) {

  mud = m;
  mud->incrRefCount(false);

  init();
  setName(m->getName());
}

Connection::Connection(int fd, MUD * m, int compress, int cookie) {
  mud = m;
  mud->incrRefCount(false);

  init();

  s->fd = fd;

  isclosed = false;
  copyover_event = FALSE;

  timeout = gtk_timeout_add(200, read_socket_timeout, (gpointer)this);

  // When doing copyover we're not too likely to call connect()
  keepalive_timer = gtk_timeout_add(60000, connection_keepalive, (gpointer)this);
  keepalive_pulses = 0;

  setName(m->getName());
}

void auto_reconnect_init(Connection *);

void Connection::closed() {
  isclosed = true;
  auto_reconnect_init(this);
}

bool Connection::isClosed() {
  return isclosed;
}

void Connection::failed() {
  char buf[2048];
#if defined(sun)
  sprintf(buf, "Connection to %.900s port %d failed: reason unknown (solaris has no sys_errlist).\n", getHost(), getPort());
#else
  sprintf(buf, "Connection to %.900s port %d failed: %s.\n", getHost(), getPort(), sys_errlist[connected * -1]);
#endif
  new Message(_("Connection failed."), buf, true);
}

Connection::~Connection() {

  if (deleted)
    return;

  deleted = TRUE;

  close();

  mainWindow->removeConnection(this);

  if (timeout)
    gtk_timeout_remove(timeout);

  if (keepalive_timer) {
    gtk_timeout_remove(keepalive_timer);
    keepalive_timer = 0;
  }

  delete history;
  history = NULL;

  delete vt;
  vt = NULL;

  delete s;
  s = NULL;

  // Save the old last MUD so we can destroy it if needed
  MUD * old = lastMUD;

  // Assign lastMUD for the reconnection function
  // We don't change the reference count, as we are
  // holding over the reference from the inital Connection
  lastMUD = mud;

  // Decrement the reference count on the old MUD and destroy if needed
  if (old) {
    old->decrRefCount(false);
    if (old->getRefCount() == 0)
      delete old;
  }

  if (name)
    free(name);

}


int Connection::getPort() {
  return mud->getPort();
}

History * Connection::getHistory() {
  return history;
}

char * Connection::getName() {
  return name;
}

void Connection::setName(char * n) {
  if (name)
    free(name);

  name = strdup(n);
}

int login_trigger(regex_t * reg, Connection * c, char * in, char * stripped_in, void * data) {
  MUD * mud = c->getMUD();
  char buf2[1024];
  sprintf(buf2, "%s\n", mud->getLoginName());
  c->getSocket()->write(buf2, strlen(buf2));
  return 1;
}

int password_trigger(regex_t * reg, Connection * c, char * in, char * stripped_in, void * data) {
  MUD * mud = c->getMUD();
  char buf2[1024];
  sprintf(buf2, "%s\n", mud->getPassword());
  c->getSocket()->write(buf2, strlen(buf2));
  return 1;
}

int Connection::connect() {
  getVT()->append("Connecting...");

  if (s->create(getHost(),getPort()) == -1) {
    connected = errno * -1;
    return -1;
  }

  isclosed = false;

  if (timeout)
    gtk_timeout_remove(timeout);

  timeout = gtk_timeout_add(200, read_socket_timeout, (gpointer)this);

  keepalive_timer = gtk_timeout_add(60000, connection_keepalive, (gpointer)this);
  keepalive_pulses = 0;

  if (strlen(mud->getLoginTrigger()) > 0 && strlen(mud->getLoginName()) > 0) {
    loginTrigger = new SystemTriggerEntity(mud->getLoginTrigger(), this, &login_trigger, NULL);
    entities->addEntity("Login Trigger", loginTrigger);
  }

  if (strlen(mud->getPasswordTrigger()) > 0 && strlen(mud->getPassword()) > 0) {
    passwordTrigger = new SystemTriggerEntity(mud->getPasswordTrigger(), this, &password_trigger, NULL);
    entities->addEntity("Login Trigger", passwordTrigger);
  }

  return 1;
}

int Connection::reconnect() {
  int res;

  // Reset VT session state settings
  getVT()->showCommands();
  getVT()->setInput("");

  // Disconnect the socket.
  close();

  // Delete the socket.
  delete s;

  if (queryPreferences()->getHttpConnect()) {
    s = new SquidSocket(this);
  } else {
    if (queryPreferences()->getTurfHttpd()) {
      s = new HTTPSocket(this);
    } else {
      if (queryPreferences()->getSocks()) {
	if (queryPreferences()->getSocks4()) {
	  s = new Socks(this);
	} else {
	  s = new SocksFive(this);
	}
      } else {
	s = new Socket(this);
      }
    }
  }

  // Reconnect
  res = connect();
  return res;
}

VT * Connection::getVT() {
  return vt;
}

Socket * Connection::getSocket() {
  return s;
}

char * Connection::getHost() {
  return mud->getHostname();
}

Connection * Connection::getNext() {
  return next;
}

void Connection::setNext(Connection * c) {
  next = c;
}

int Connection::isConnected() {

  if (connected)
    return 1;

  // No socket!
  if (!s)
    return -1;

  if (s->connected() == -1) {
    connected = errno * -1;
    return -1;
  }

  if (s->connected() == 1) {
    connected = 1;
    Event * e = new Event(EvConnect, NULL);
    phandler->fireEvent(e, this);
    delete e;
    if (copyover_event)
      getVT()->append("  Connected.\n");
  } else 
    getVT()->append(".");

  return connected;
}

void Connection::close() {

  if (loginTrigger) {
    entities->removeEntity(loginTrigger);
    delete loginTrigger;
    loginTrigger = NULL;
  }

  if (passwordTrigger) {
    entities->removeEntity(passwordTrigger);
    delete passwordTrigger;
    passwordTrigger = NULL;
  }

  if (connected > 0) {
    Event * e = new Event(EvDisconnect, NULL);
    phandler->fireEvent(e, this);
    delete e;
  }

  if (keepalive_timer) {
    gtk_timeout_remove(keepalive_timer);
    keepalive_timer = 0;
  }

  if (timeout) {
    gtk_timeout_remove(timeout);
    timeout = 0;
  }

  if (s)
    s->dispose();

  connected = 0;
}

void Connection::setCharMode(bool b) {

  if (!queryPreferences()->getNegotiateCharmode())
    return;

  if (b == charMode)
    return;

  charMode = b;

  if (charMode)
    getVT()->setMode(VTCharacterMode);
  else
    getVT()->setMode(VTLineMode);

}

bool Connection::getCharMode() {
  return charMode;
}

MUD * Connection::getMUD() {
  return mud;
}

/**
 * Returns the preferences object for this connection.  NULL if it is
 * using the default preferences.
 */

Prefs * Connection::getPreferences() {
  return mud->getPreferences();
}

Prefs * Connection::queryPreferences() {
  if (getPreferences())
    return getPreferences();
  
  return globalPreferences;
}

gint connection_keepalive(gpointer data) {
  Connection * c = (Connection *) data;

  // See if we should send data this minute - this goes above
  // the de-activation check so that it will work properly when
  // first set by the user during a run
  if (++c->keepalive_pulses < c->queryPreferences()->getKeepaliveTimer())
    return TRUE;

  // 0 or -1 both isable the feature
  if (c->queryPreferences()->getSendKeepalive() == FALSE
      || c->queryPreferences()->getKeepaliveTimer() <= 0)
    return TRUE;

  // Send some bytes that most muds ignore for the most part
  ::keepalive_send(c);
  c->keepalive_pulses = 0;

  return TRUE;
}

void Connection::keepalive_send(Connection * c) {
  c->getSocket()->write("\n\r\n\r");
}

MUD * Connection::getLastMUD() {
  return lastMUD;
}
